#include <LiquidCrystal.h>
#include <Wire.h>
#include <SPI.h>
#include <Adafruit_PN532.h>
#include <SoftwareSerial.h>
#include <DFRobotDFPlayerMini.h>

#include "station.h"
#include "statemachine.h"
#include "state.h"
#include "states.h"
#include "timer.h"
#include "audioState.h"

#define PN532_CS 10

// Using this library to set a timeout on readPassiveTagID
// The library in use is slightly modified to fix this issue: https://github.com/adafruit/Adafruit-PN532/issues/117
Adafruit_PN532 nfc(PN532_CS);

constexpr int rs = 9, en = 8, d4 = 5, d5 = 4, d6 = 3, d7 = 2;
LiquidCrystal lcd(rs, en, d4, d5, d6, d7);

SoftwareSerial softSerial(6, 7); // RX, TX
DFRobotDFPlayerMini dfPlayer;

Station stations[5] = {
    Station(1680767519, "Yellow"), Station(3346823711, "Green"), Station(3569318175, "Pink"),
    Station(2174777887, "Blue"), Station(102951984, "Weird")
};
Station* currentStation;
int currentStationIndex = -1;

Timer stationDetectionTimer(500, true);

String previousDisplayContent;

StateMachine stateMachine(new Startup());

unsigned long lastLoopTime = 0;
unsigned long currentLoopTime = 0;
unsigned long deltaTime = 0;


void setupNFC()
{
    nfc.begin();

    uint32_t versiondata = nfc.getFirmwareVersion();
    if (!versiondata)
    {
        while (1); // halt
    }
    // configure board to read RFID tags and cards
    nfc.SAMConfig();
}


Station* detectStation()
{
    uint8_t uid[7];
    uint8_t uidLength;
    constexpr uint16_t timeout = 100;
    const bool success = nfc.readPassiveTargetID(PN532_MIFARE_ISO14443A, uid, &uidLength, timeout);
    if (!success)
    {
        return nullptr;
    }

    uint32_t id = 0;
    for (uint8_t i = 0; i < uidLength; i++)
    {
        id <<= 8;
        id |= uid[i];
    }

    Station* found = 0;
    for (unsigned int i = 0; i < sizeof(stations) / sizeof(stations[0]); i++)
    {
        Station* currentStation = &stations[i];
        if (currentStation->check(id))
        {
            found = currentStation;
        }
    }
    return found;
}

void checkStations()
{
    Station* station = detectStation();

    if (station != nullptr)
    {
        if (currentStation == nullptr)
        {
            currentStation = station;
            stateMachine.putDown(Context(stations, sizeof(stations) / sizeof(stations[0]), deltaTime, &dfPlayer), currentStation);
        }
    }
    else
    {
        if (currentStation != nullptr)
        {
            currentStation = nullptr;
            stateMachine.pickedUp(Context(stations, sizeof(stations) / sizeof(stations[0]), deltaTime, &dfPlayer));
        }
    }
}

void printString(const String& str)
{
    int currentColumn = 0;
    int currentRow = 0;
    lcd.clear();

    for (unsigned int i = 0; i < str.length(); i++)
    {
        lcd.setCursor(currentColumn, currentRow);
        lcd.print(str[i]);
        currentColumn++;
        if (currentColumn >= 16)
        {
            currentColumn = 0;
            currentRow++;
            if (currentRow >= 2)
            {
                break;
            }
        }
    }
}

void loop(void)
{
    currentLoopTime = millis();
    deltaTime = currentLoopTime - lastLoopTime;

    bool doCheck = stationDetectionTimer.update(deltaTime);
    if (doCheck)
    {
        checkStations();
    }

    stateMachine.update(Context(stations, sizeof(stations) / sizeof(stations[0]), deltaTime, &dfPlayer));
    String newContent = stateMachine.updateDisplay();
    if (previousDisplayContent != newContent)
    {
        lcd.clear();
        printString(newContent);
        previousDisplayContent = newContent;
    }

    lastLoopTime = currentLoopTime;
    delay(10);
}

void setup(void)
{
    setupNFC();

    Serial.begin(9600);
    
    lcd.begin(16, 2);
    stationDetectionTimer.start();

    softSerial.begin(9600);

    lcd.print("Initializing MP3");
    lcd.clear();
    if (!dfPlayer.begin(softSerial, true, true))
    {
        lcd.print("Failed to init");
        lcd.setCursor(0, 1);
        lcd.print("DFPlayer!");
        while (true)
        {
            delay(100);
            printMp3Detail(&lcd, dfPlayer.readType(), dfPlayer.read());
        }
    }

    // dfPlayer.volume(15);
    // dfPlayer.play(1);
}
